
package com.knownspace.tools;

import com.knownspace.tools.MalformedPatternException;

public class StringSearcher
	{
	/*
	Search a target string for a pattern string using a simplified
	Boyer-Moore algorithm.

	Public Methods and Constructors:
	--------------------------------
	StringSearcher(String pattern)
		initialize a new StringSearcher to search for the given pattern

	setPattern(String pattern)
		reset the pattern to be used in subsequent searches

	int find(String target)
		return the index of the first character in the target string
		matching the pattern, otherwise return PATTERN_NOT_FOUND

	int find(String target, int startPosition)
		return the index of the first character in the target string
		matching the pattern starting at startPosition, otherwise return
		PATTERN_NOT_FOUND

	Comments:
	---------
	Boyer-Moore tries to match the pattern against the target from left to
	right, but at each search position it tries to match the pattern from
	right to left.
	--If the first comparison is a mismatch and the target character does not
	appear in the pattern, right shift the pattern by its length.
	--If the first comparison is a mismatch but the target character appears
	somewhere in the pattern, right shift the pattern to lineup the rightmost
	appearance of that pattern character with the target character.
	--If the first comparison is a match, test the next rightmost character in
	the pattern against the next rightmost character in the target.
	--If the last comparison was a match and the pattern's leftmost character
	was just tested, then the pattern has been found.

	There is a more complex, and more efficient, version of Boyer-Moore not
	implemented here.

	Class Relationships:
	--------------------
	This class can throw objects of class MalformedPatternException
	*/

	//the search pattern
	private String pattern;

	//the jump table based on characters in the pattern
	private int[] characterJump;

	//the characterJump table must hold every possible target character
	//(for unicode use 65536; for 8-bit ascii use 256; for 7-bit ascii use 128)
	private static final int ALPHABET_SIZE = 128;

	public static final int PATTERN_NOT_FOUND = -1;

	/////////////////////////////////////////////////////////////////////////
	public StringSearcher(final String pattern)
		throws MalformedPatternException
		{
		/*
		Create a new string searcher for the given pattern.

		Throw a MalformedPatternException if the pattern is null or empty.
		*/

		setPattern(pattern);
		}

	/////////////////////////////////////////////////////////////////////////
	public void setPattern(final String pattern)
		throws MalformedPatternException
		{
		/*
		Save the pattern and create the corresponding Boyer-Moore table.

		Throw a MalformedPatternException if the pattern is null or empty.
		*/

		if ((pattern == null) || (pattern.length() == 0))
			throw new MalformedPatternException();

		this.pattern = pattern;

		setCharacterJumps();
		}

	/////////////////////////////////////////////////////////////////////////
	private void setCharacterJumps()
		{
		/*
		Setup the table to handle character mismatch jumps.
		*/

		characterJump = new int[ALPHABET_SIZE];

		//the default jump for every character is the length of the pattern;
		//this takes care of the case where the character is not in the pattern
		for (int i = 0; i < ALPHABET_SIZE; i++)
			characterJump[i] = pattern.length();

		//the jump value for each character actually in the pattern
		//is that character's rightmost position within the pattern;
		//this makes the two characters lineup for the next test
		for (int i = 0; i < pattern.length(); i++)
			characterJump[pattern.charAt(i)] = pattern.length() - 1 - i;
		}

	/////////////////////////////////////////////////////////////////////////
	public int find(final String target)
		{
		/*
		Search all of the target for the first occurrence of the pattern.
		If the pattern is found, return the index of the target character
		matching the first character of the pattern, otherwise return
		PATTERN_NOT_FOUND.
		*/

		return find(target, 0);
		}

	/////////////////////////////////////////////////////////////////////////
	public int find(final String target, final int startPosition)
		{
		/*
		Search the target for the first occurrence of the pattern starting
		from startPosition on. If the pattern is found, return the index of
		the target character matching the first character of the pattern,
		otherwise return PATTERN_NOT_FOUND.
		*/

		//make sure the arguments are sane
		if ((startPosition < 0) ||
			(startPosition > target.length() - pattern.length()))
			return PATTERN_NOT_FOUND;

		int targetIndex = startPosition + pattern.length() - 1;
		int patternIndex = pattern.length() - 1;
		while (targetIndex < target.length())
			{
			if (pattern.charAt(patternIndex) == target.charAt(targetIndex))
				{
				//characters match; have we checked all of the pattern?
				if (patternIndex == 0)
					{
					//pattern exhausted; pattern found
					return targetIndex;
					}
				else
					{
					//check next rightmost characters
					targetIndex--;
					patternIndex--;
					}
				}
			else
				{
				//characters mismatch; jump target forward and reset pattern
				targetIndex += characterJump[target.charAt(targetIndex)];
				patternIndex = pattern.length() - 1;
				}
			}

		return PATTERN_NOT_FOUND;
		}

	/////////////////////////////////////////////////////////////////////////
	public static void main(final String[] commandLineArguments)
		throws MalformedPatternException
		{
		/*
		Exercise the above code.
		*/

		StringSearcher searcher = new StringSearcher("wowwow");

		int found = searcher.find("how are you? wowwow hello");

		if (found == PATTERN_NOT_FOUND)
			System.out.println("pattern not found");
		else
			System.out.println("pattern found at location: " + found);

		found = searcher.find("how are you? owwow hello");

		if (found == PATTERN_NOT_FOUND)
			System.out.println("pattern not found");
		else
			System.out.println("pattern found at location: " + found);

		searcher.setPattern("<");
		found = searcher.find("how are <you? owwow < hello");

		if (found == PATTERN_NOT_FOUND)
			System.out.println("pattern not found");
		else
			System.out.println("pattern found at location: " + found);

		found = searcher.find("how are <you? owwow < hello", found + 1);

		if (found == PATTERN_NOT_FOUND)
			System.out.println("pattern not found");
		else
			System.out.println("pattern found at location: " + found);
		}
	}
